@@ -23,7 +23,8 @@ from threading import Timer as _Timer
 
 from gi.repository import Gdk, GLib, Gtk
 from logitech_receiver.settings import KIND as _SETTING_KIND
-from solaar.i18n import _
+from logitech_receiver.settings import SENSITIVITY_IGNORE as _SENSITIVITY_IGNORE
+from solaar.i18n import _, ngettext
 from solaar.ui import ui_async as _ui_async
 
 #
@@ -31,12 +32,12 @@ from solaar.ui import ui_async as _ui_async
 #
 
 
-def _read_async(setting, force_read, sbox, device_is_online):
-    def _do_read(s, force, sb, online):
+def _read_async(setting, force_read, sbox, device_is_online, sensitive):
+    def _do_read(s, force, sb, online, sensitive):
         v = s.read(not force)
-        GLib.idle_add(_update_setting_item, sb, v, online, priority=99)
+        GLib.idle_add(_update_setting_item, sb, v, online, sensitive, priority=99)
 
-    _ui_async(_do_read, setting, force_read, sbox, device_is_online)
+    _ui_async(_do_read, setting, force_read, sbox, device_is_online, sensitive)
 
 
 def _write_async(setting, value, sbox):
@@ -104,7 +105,7 @@ def _create_choice_control(setting):
     c = Gtk.ComboBoxText()
     # TODO i18n text entries
     for entry in setting.choices:
-        c.append(str(int(entry)), str(entry))
+        c.append(str(int(entry)), _(str(entry)))
     c.connect('changed', _combo_notify, setting)
     return c
 
@@ -132,7 +133,7 @@ def _create_map_choice_control(setting):
         if choices:
             # TODO i18n text entries
             for choice in choices:
-                valueBox.append(str(int(choice)), str(choice))
+                valueBox.append(str(int(choice)), _(str(choice)))
             if current is not None:
                 valueBox.set_active_id(str(int(current)))
 
@@ -143,7 +144,7 @@ def _create_map_choice_control(setting):
     c.pack_end(valueBox, False, False, 0)
     # TODO i18n text entries
     for entry in setting.choices:
-        keyBox.append(str(int(entry)), str(entry))
+        keyBox.append(str(int(entry)), _(str(entry)))
     keyBox.set_active(0)
     keyBox.connect('changed', _map_value_notify_key, (setting, valueBox))
     _map_populate_value_box(valueBox, setting, int(keyBox.get_active_id()))
@@ -181,7 +182,7 @@ def _create_slider_control(setting):
     return control.gtk_range
 
 
-def _create_multiple_toggle_control(setting):
+def _create_multiple_toggle_control(setting, change):
     def _toggle_notify(control, _, setting):
         if control.get_sensitive():
             key = control._setting_key
@@ -189,7 +190,7 @@ def _create_multiple_toggle_control(setting):
             if setting._value[key] != new_state:
                 setting._value[key] = new_state
                 p = control
-                for _ in range(5):
+                for _ in range(5):  # go up widget chain
                     p = p.get_parent()
                 _write_async_key_value(setting, key, new_state, p)
 
@@ -209,6 +210,7 @@ def _create_multiple_toggle_control(setting):
     lb._showing = True
     lb.set_selection_mode(Gtk.SelectionMode.NONE)
     lb.set_no_show_all(True)
+    lb._label_control_pairs = []
     btn = Gtk.Button('? / ?')
     for k in setting._validator.all_options():
         h = Gtk.HBox(homogeneous=False, spacing=0)
@@ -230,20 +232,23 @@ def _create_multiple_toggle_control(setting):
         lbl.set_alignment(0.0, 0.5)
         lbl.set_margin_left(30)
         lb.add(h)
+        lb._label_control_pairs.append((lbl, control))
     _disable_listbox_highlight_bg(lb)
     lb._toggle_display()
     btn.connect('clicked', lambda _: lb._toggle_display())
 
-    hbox = Gtk.HBox(homogeneous=False, spacing=60)
+    hbox = Gtk.HBox(homogeneous=False, spacing=6)
+    hbox.pack_end(change, False, False, 0)
     hbox.pack_end(btn, False, False, 0)
     btn.set_alignment(1.0, 0.5)
     vbox = Gtk.VBox(homogeneous=False, spacing=6)
     vbox.pack_start(hbox, True, True, 0)
     vbox.pack_end(lb, True, True, 0)
+    vbox._header, vbox._button, vbox._control = hbox, btn, lb
     return vbox
 
 
-def _create_multiple_range_control(setting):
+def _create_multiple_range_control(setting, change):
     def _write(control, setting, item, sub_item):
         control._timer.cancel()
         delattr(control, '_timer')
@@ -278,6 +283,7 @@ def _create_multiple_range_control(setting):
     lb.set_selection_mode(Gtk.SelectionMode.NONE)
     lb._showing = True
     lb.set_no_show_all(True)
+    lb._items = []
     btn = Gtk.Button('...')
     for item in setting._validator.items:
         lbl_text = str(item)
@@ -293,6 +299,7 @@ def _create_multiple_range_control(setting):
         lb.set_tooltip_text(lbl_tooltip or ' ')
         item_lb = Gtk.ListBox()
         item_lb.set_selection_mode(Gtk.SelectionMode.NONE)
+        item_lb._sub_items = []
         for sub_item in setting._validator.sub_items[item]:
             h = Gtk.HBox(homogeneous=False, spacing=20)
             lbl_text = str(sub_item)
@@ -322,18 +329,23 @@ def _create_multiple_range_control(setting):
             control.connect('value-changed', _changed, setting, item, sub_item)
             item_lb.add(h)
             h._setting_sub_item = sub_item
+            h._label, h._control = sub_item_lbl, control
+            item_lb._sub_items.append(h)
         item_lb._setting_item = item
         _disable_listbox_highlight_bg(item_lb)
         lb.add(item_lb)
+        lb._items.append(item_lb)
     _disable_listbox_highlight_bg(lb)
     lb._toggle_display()
     btn.connect('clicked', lambda _: lb._toggle_display())
     btn.set_alignment(1.0, 0.5)
     hbox = Gtk.HBox(homogeneous=False, spacing=6)
+    hbox.pack_end(change, False, False, 0)
     hbox.pack_end(btn, False, False, 0)
     vbox = Gtk.VBox(homogeneous=False, spacing=6)
     vbox.pack_start(hbox, True, True, 0)
     vbox.pack_end(lb, True, True, 0)
+    vbox._header, vbox._button, vbox._control = hbox, btn, lb
     return vbox
 
 
@@ -341,11 +353,49 @@ def _create_multiple_range_control(setting):
 #
 #
 
+_allowables_icons = {True: 'changes-allow', False: 'changes-prevent', _SENSITIVITY_IGNORE: 'dialog-error'}
+_allowables_tooltips = {
+    True: _('Changes allowed'),
+    False: _('No changes allowed'),
+    _SENSITIVITY_IGNORE: _('Ignore this setting')
+}
+_next_allowable = {True: False, False: _SENSITIVITY_IGNORE, _SENSITIVITY_IGNORE: True}
+_icons_allowables = {v: k for k, v in _allowables_icons.items()}
 
-def _create_sbox(s):
+
+# clicking on the lock icon changes from changeable to unchangeable to ignore
+def _change_click(eb, button, arg):
+    control, sbox, device, name = arg
+    icon = eb.get_children()[0]
+    icon_name, _ = icon.get_icon_name()
+    allowed = _icons_allowables.get(icon_name, True)
+    new_allowed = _next_allowable[allowed]
+    control.set_sensitive(new_allowed is True)
+    _change_icon(new_allowed, icon)
+    if device.persister:  # remember the new setting sensitivity
+        device.persister.set_sensitivity(name, new_allowed)
+    if allowed == _SENSITIVITY_IGNORE:  # update setting if it was being ignored
+        setting = next((s for s in device.settings if s.name == name), None)
+        if setting:
+            persisted = device.persister.get(setting.name) if device.persister else None
+            if persisted is not None:
+                _write_async(setting, persisted, sbox)
+            else:
+                _read_async(setting, True, sbox, bool(device.online), control.get_sensitive())
+    return True
+
+
+def _change_icon(allowed, icon):
+    allowed = allowed if allowed in _allowables_icons else True
+    icon.set_from_icon_name(_allowables_icons[allowed], Gtk.IconSize.LARGE_TOOLBAR)
+    icon.set_tooltip_text(_allowables_tooltips[allowed])
+
+
+def _create_sbox(s, device):
     sbox = Gtk.HBox(homogeneous=False, spacing=6)
-    label = Gtk.Label(s.label)
-    sbox.pack_start(label, False, False, 0)
+    lbl = Gtk.Label(s.label)
+    label = Gtk.EventBox()
+    label.add(lbl)
 
     spinner = Gtk.Spinner()
     spinner.set_tooltip_text(_('Working') + '...')
@@ -353,43 +403,55 @@ def _create_sbox(s):
     failed = Gtk.Image.new_from_icon_name('dialog-warning', Gtk.IconSize.SMALL_TOOLBAR)
     failed.set_tooltip_text(_('Read/write operation failed.'))
 
+    change_icon = Gtk.Image.new_from_icon_name('changes-prevent', Gtk.IconSize.LARGE_TOOLBAR)
+    _change_icon(False, change_icon)
+    change = Gtk.EventBox()
+    change.add(change_icon)
+
     if s.kind == _SETTING_KIND.toggle:
         control = _create_toggle_control(s)
-        sbox.pack_end(control, False, False, 0)
     elif s.kind == _SETTING_KIND.choice:
         control = _create_choice_control(s)
-        sbox.pack_end(control, False, False, 0)
     elif s.kind == _SETTING_KIND.range:
         control = _create_slider_control(s)
-        sbox.pack_end(control, True, True, 0)
     elif s.kind == _SETTING_KIND.map_choice:
         control = _create_map_choice_control(s)
-        sbox.pack_end(control, True, True, 0)
     elif s.kind == _SETTING_KIND.multiple_toggle:
-        vbox = _create_multiple_toggle_control(s)
-        control = vbox.get_children()[1]
-        sbox.remove(label)
-        vbox.get_children()[0].pack_start(label, False, False, 0)
-        label.set_alignment(0.0, 0.5)
+        vbox = _create_multiple_toggle_control(s, change)
+        control = vbox._control
+        lbl.set_alignment(0.0, 0.5)
         sbox.pack_start(vbox, True, True, 0)
     elif s.kind == _SETTING_KIND.multiple_range:
-        vbox = _create_multiple_range_control(s)
-        control = vbox.get_children()[1]
-        sbox.remove(label)
-        vbox.get_children()[0].pack_start(label, False, False, 0)
-        label.set_alignment(0.0, 0.5)
+        vbox = _create_multiple_range_control(s, change)
+        control = vbox._control
+        lbl.set_alignment(0.0, 0.5)
         sbox.pack_start(vbox, True, True, 0)
     else:
         raise Exception('NotImplemented')
     control.set_sensitive(False)  # the first read will enable it
     control.kind = s.kind
 
+    change.set_sensitive(True)
+    change.connect('button-press-event', _change_click, (control, sbox, device, s.name))
+
     if s.kind in [_SETTING_KIND.multiple_toggle, _SETTING_KIND.multiple_range]:
-        vbox.get_children()[0].pack_end(spinner, False, False, 0)
-        vbox.get_children()[0].pack_end(failed, False, False, 0)
+        vbox._header.pack_start(label, False, False, 0)
+        vbox._header.pack_end(spinner, False, False, 0)
+        vbox._header.pack_end(failed, False, False, 0)
+        sbox._button = vbox._button
     else:
+        sbox.pack_start(label, False, False, 0)
+        sbox.pack_end(change, False, False, 0)
+        sbox.pack_end(control, s.kind == _SETTING_KIND.range, s.kind == _SETTING_KIND.range, 0)
         sbox.pack_end(spinner, False, False, 0)
         sbox.pack_end(failed, False, False, 0)
+    sbox._label = label
+    sbox._lbl = lbl
+    sbox._spinner = spinner
+    sbox._failed = failed
+    sbox._change = change
+    sbox._change_icon = change_icon
+    sbox._control = control
 
     if s.description:
         sbox.set_tooltip_text(s.description)
@@ -402,13 +464,14 @@ def _create_sbox(s):
     return sbox
 
 
-def _update_setting_item(sbox, value, is_online=True):
+def _update_setting_item(sbox, value, is_online=True, sensitive=True):
     failed, spinner, control = _get_failed_spinner_control(sbox)
     spinner.set_visible(False)
     spinner.stop()
 
     if value is None:
         control.set_sensitive(False)
+        _change_icon(False, sbox._change_icon)
         failed.set_visible(is_online)
         return
 
@@ -422,64 +485,54 @@ def _update_setting_item(sbox, value, is_online=True):
         control.set_value(int(value))
     elif isinstance(control, Gtk.HBox):
         kbox, vbox = control.get_children()  # depends on box layout
-        if value.get(kbox.get_active_id()):
+        if value.get(kbox.get_active_id()) is not None:
             vbox.set_active_id(str(value.get(kbox.get_active_id())))
     elif isinstance(control, Gtk.ListBox):
         if control.kind == _SETTING_KIND.multiple_toggle:
-            total = len(control.get_children())
+            total = len(control._label_control_pairs)
             active = 0
             to_join = []
-            for ch in control.get_children():
-                elem = ch.get_children()[0].get_children()[-1]
+            for lbl, elem in control._label_control_pairs:
                 v = value.get(elem._setting_key, None)
                 if v is not None:
                     elem.set_active(v)
                 if elem.get_active():
                     active += 1
-                to_join.append(elem.get_parent().get_children()[0].get_text() + ': ' + str(elem.get_active()))
+                to_join.append(lbl.get_text() + ': ' + str(elem.get_active()))
             b = ', '.join(to_join)
-            btn = control.get_parent().get_children()[0].get_children()[-1]
-            btn.set_label(f'{active} / {total}')
-            btn.set_tooltip_text(b)
+            sbox._button.set_label(f'{active} / {total}')
+            sbox._button.set_tooltip_text(b)
         elif control.kind == _SETTING_KIND.multiple_range:
             b = ''
             n = 0
-            for ch in control.get_children()[1:]:
+            for ch in control._items:
                 # item
-                item = ch.get_children()[0]._setting_item
+                item = ch._setting_item
                 v = value.get(str(int(item)), None)
                 if v is not None:
                     b += str(item) + ': ('
                     to_join = []
-                    for c in ch.get_children()[0].get_children():
+                    for c in ch._sub_items:
                         # sub-item
-                        row = c.get_children()[0]
-                        sub_item = row._setting_sub_item
-                        elem = row.get_children()[-1]
-                        elem.set_value(v[str(sub_item)])
+                        sub_item = c._setting_sub_item
+                        c._control.set_value(v[str(sub_item)])
                         n += 1
                         to_join.append(str(sub_item) + f'={v[str(sub_item)]}')
                     b += ', '.join(to_join) + ') '
-                btn = control.get_parent().get_children()[0].get_children()[-1]
-                btn.set_label(f'{n} value' + ('s' if n != 1 else ''))  # TODO: i18n, singular/plural
-                btn.set_tooltip_text(b)
+                lbl_text = ngettext('%d value', '%d values', n) % n
+                sbox._button.set_label(lbl_text)
+                sbox._button.set_tooltip_text(b)
         else:
             raise NotImplementedError
     else:
         raise Exception('NotImplemented')
-    control.set_sensitive(True)
+
+    control.set_sensitive(sensitive is True)
+    _change_icon(sensitive, sbox._change_icon)
 
 
 def _get_failed_spinner_control(sbox):
-    children = sbox.get_children()
-    if len(children) == 4:
-        _ignore, failed, spinner, control = sbox.get_children()  # depends on box layout
-    else:
-        assert len(children) == 1
-        control = children[0].get_children()[-1]
-        failed = children[0].get_children()[0].get_children()[1]
-        spinner = children[0].get_children()[0].get_children()[2]
-    return failed, spinner, control
+    return sbox._failed, sbox._spinner, sbox._control
 
 
 def _disable_listbox_highlight_bg(lb):
@@ -528,10 +581,11 @@ def update(device, is_online=None):
         if k in _items:
             sbox = _items[k]
         else:
-            sbox = _items[k] = _create_sbox(s)
+            sbox = _items[k] = _create_sbox(s, device)
             _box.pack_start(sbox, False, False, 0)
 
-        _read_async(s, False, sbox, is_online)
+        sensitive = device.persister.get_sensitivity(s.name) if device.persister else True
+        _read_async(s, False, sbox, is_online, sensitive)
 
     _box.set_visible(True)
 
